from openai import OpenAI

client = OpenAI()

def _openai_model(model_name, prompt, seed):
    response = client.chat.completions.create(
            model=model_name,
            messages=[
                {"role": "user", "content": prompt}
            ],
            temperature=0,
            seed=seed
            )

    return response.choices[0].message.content



from tqdm import tqdm
import numpy as np
from importlib import resources

from ..utils import preprocessing

import warnings
warnings.filterwarnings('ignore')


class ICLModel():
    def __init__(
        self, 
        llm_model, 
        params
        ):
        self.llm_model = llm_model
        self.columns_name = params['columns_name']
        self.objective = params['objective']
        self.icl_prompt = None


    def fit(self, x, y):
        df = x.copy()
        df['target'] = y

        # Create a string dataset
        dataset_str = preprocessing.text_converter(df)

        # create prompt
        if prompt == None:
            if self.objective == 'regression':
                with resources.open_text('iblm.iclmodel.prompt', 'regression.txt') as file:
                    prompt = file.read()
            elif self.objective == 'classification':
                with resources.open_text('iblm.iclmodel.prompt', 'classification.txt') as file:
                    prompt = file.read()

        icl_prompt = prompt.format(dataset_str_ = dataset_str)
            
        self.icl_prompt = icl_prompt

        num_prompt = len(icl_prompt)
        print(f'Number of input tokens:{num_prompt}')

        return icl_prompt



    def predict(self, x):
        if self.icl_prompt is None:
            raise Exception("You must train the model before predicting!")

        output = []
        for _, row in tqdm(x.iterrows(), total=x.shape[0]):
            str_row = ','.join([str(elm) for elm in row.to_list()])
            prompt = self.icl_prompt + str_row

            while True:
                try:
                    y = self.llm_model(prompt)
                    y = float(y)
                    output.append(y)
                    break  
                except ValueError:
                    pass

        return np.array(output)






import pandas as pd
from openai import OpenAI

client = OpenAI()

def _data_to_text(X, y):
    df = X.copy()
    df['target'] = y

    data_type = ', '.join(df.dtypes.astype(str))

    dataset = []
    for index, row in df.iterrows():
        row_as_str = [str(item) for item in row.tolist()]
        dataset.append(','.join(row_as_str))
    dataset_str = '\n'.join(dataset)
    return dataset_str, data_type


def columns_name(columns_name, df):
    if columns_name:
        col_name = ', '.join(df.columns.astype(str))
        col_option = ''
    else:
        # serial number
        df.columns = range(df.shape[1])
        col_name = ', '.join(df.columns.astype(str))
        col_option = 'df.columns = range(df.shape[1])'
    return col_name, col_option


def _prompt_modification(code_model):
    modification_prompt = """
        Please extract and output only the Python code from the following.
        Do not put ```python ```, etc. into the output.
        -------------
        {code_model_}
        """.format(code_model_ = code_model)
    return modification_prompt


def _save_codemodel(file_path, file_name, code_model):
    if file_path != None:
        with open(file_path + f'{file_name}.py', mode='w') as file:
            file.write(code_model)
    return


def _interpret_codemodel(code_model):
    interpret_prompt = """
    Refer to the code below and explain how you are going to process the data and make predictions.
    The only part to explain is the part where the data is processed.
    Do not explain df = x.copy().
    Please output the data in bulleted form.
    Please tell us what you can say based on the whole process.
    ------------------
    {code_model_}
    """.format(code_model_ = code_model)

    return interpret_prompt



import re

import numpy as np

import warnings
warnings.filterwarnings('ignore')



class IBLBoosting():
    def __init__(
        self,
        llm_model_name,
        params
        ):
        self.llm_model_name = llm_model_name
        self.llm_model = OpenAI(temperature=0, model_name = self.llm_model_name)

        #self.llm_model = llm_model,
        self.columns_name = params['columns_name']
        self.model_code = None


    def fit(self, x, y, model_name, file_path=None):
        print("> Start of model creating.")
        df = x.copy()

        df['target'] = y

        # Determine whether binary or multivalued classification is used
        if len(df['target'].unique()) == 2:
            task_type = 'binary classification'
            output_code = 'y = 1 / (1 + np.exp(-y))'
        else:
            task_type = 'multi-class classification'

        # Obtaining data types
        data_type = ', '.join(df.dtypes.astype(str))



        # Create a string dataset
        dataset = []
        for index, row in df.iterrows():
            row_as_str = [str(item) for item in row.tolist()]
            dataset.append(','.join(row_as_str))
        dataset_str = '\n'.join(dataset)


        # column name
        if self.columns_name:
            col_name = ', '.join(df.columns.astype(str))
            col_option = ''

        else:
            # serial number
            df.columns = range(df.shape[1])
            col_name = ', '.join(df.columns.astype(str))
            col_option = 'df.columns = range(df.shape[1])'



        create_prompt = """
        Please create your code in compliance with all of the following conditions. Output should be code only. Do not enclose the output in ``python ``` or the like.
        ・Analyze the large amount of data below and create a {task_type_} code to accurately predict "target".
        ------------------
        {dataset_str_}
        ------------------
        ・Each data type is as follows. If necessary, you can change the data type.
        ・Create code that can make predictions about new data based on logic from large amounts of input data without using machine learning models.
        ・If input is available, the column names below should also be used to help make decisions when creating the predictive model. Column Name:{col_name_}
        ・Create a code like the following. Do not change the input or output format.
        ・If {col_option_} is not blank, add it after 'df = x.copy()'.
        ・You do not need to provide examples.
        ------------------
        import numpy as np

        def predict(x):
            df = x.copy()

            output = []
            for index, row in df.iterrows():


                # Feature creation and data preprocessing


                {output_code_}
                output.append(y)

            output = np.array(output)

            return output
        """.format(
            task_type_ = task_type,
            dataset_str_ = dataset_str,
            model_name_ = model_name,
            col_name_ = col_name,
            col_option_ = col_option,
            output_code_ = output_code
            )

        #print(create_prompt)

        with get_openai_callback() as cb:
            model_code = self.llm_model(create_prompt)
            print(cb)


        # Save to File
        if file_path != None:
            with open(file_path + f'{model_name}.py', mode='w') as file:
                file.write(model_code)


        self.model_code = model_code

        return model_code

    def predict(self, x):
        if self.model_code is None:
            raise Exception("You must train the model before predicting!")

        code = self.model_code

        # = re.search(r'def (\w+)', function_string).group(1)
        #code = self.model_code + '\n'# + f'model = model({x})'
        exec(code, globals())

        #model = namespace["code"]

        y = predict(x)

        return y




    def interpret(self):
        if self.model_code is None:
            raise Exception("You must train the model before interpreting!")

        interpret_prompt = """
        Refer to the code below and explain how you are going to process the data and make predictions.
        The only part to explain is the part where the data is processed.
        Do not explain df = x.copy().
        Please output the data in bulleted form.
        Please tell us what you can say based on the whole process.
        ------------------
        {model_code_}
        """.format(
            model_code_ = self.model_code
        )

        with get_openai_callback() as cb:
            output = self.llm_model(interpret_prompt)
            print(cb)


        return output



from __future__ import annotations

import logging

from typing import TYPE_CHECKING

from .metrics import evaluate
from .prompt import make_prompt, data_to_text

from .exceptions import InvalidCodeModelError, InvalidModelObjectiveError, UndefinedCodeModelError
from .llm_client import get_client, run_prompt

import importlib.resources as pkg_resources

import re

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd


logger = logging.getLogger(__name__)
# logging.basicConfig(format="%(asctime)s [%(name)s][%(levelname)s] (%(module)s:%(filename)s")
logging.basicConfig(format="%(asctime)s [%(name)s][%(levelname)s] (%(module)s:%(filename)s:%(funcName)s:%(lineno)d)")
logger.setLevel(logging.INFO)


class IBLModel:
    IBL_OBJECTIVES = ("regression", "binary", "multiclass")

    def __init__(
        self,
        model_name: str,
        objective: str,
        # common
        api_type: str = "openai",
        # openai & azure
        api_key: str | None = None,
        max_retries: int = 5,
        timeout: int = 120,
        organization: str | None = None,
        # azure
        api_version: str | None = None,
        azure_endpoint: str | None = None,
    ) -> None:

        self.model_name = model_name
        self.objective = objective
        self.client = get_client(
            # common
            api_type=api_type,
            # openai & azure
            api_key=api_key,
            max_retries=max_retries,
            timeout=timeout,
            organization=organization,
            # azure
            api_version=api_version,
            azure_endpoint=azure_endpoint,
        )

        self.load_prompt_templates(objective)

        self.code_model = None
        self.interpret_result = None

        self.fit_params = None

    def load_prompt_templates(self, objective: str) -> None:
        model_prompt_mapping = {
            "gpt-4-0125-preview": "iblm.prompt_templates.gpt-4-0125-preview.ibl",
            "gemini-pro": "iblm.prompt_templates.gemini-pro.ibl"
        }

        task_prompt_mapping = {
            "regression": "regression.j2",
            "binary": "binary.j2",
            "multiclass": "binary.j2", # TODO: change after multiclass.j2
            "interpret": "interpret.j2",
        }

        # fit_prompt_templates
        with pkg_resources.open_text(model_prompt_mapping[self.model_name], task_prompt_mapping[self.objective]) as file:
            self._default_fit_prompt_template = file.read()

        # interpret_prompt_templates
        with pkg_resources.open_text(model_prompt_mapping[self.model_name], task_prompt_mapping["interpret"]) as file:
            self._default_interpret_prompt_template = file.read()

    @property
    def objective(self) -> str:
        return self._objective

    @objective.setter
    def objective(self, objective) -> None:
        if objective in self.IBL_OBJECTIVES:
            self._objective = objective
        else:
            raise InvalidModelObjectiveError(f"specify the objective from {self.IBL_OBJECTIVES}")

        self.load_prompt_templates(objective)

    @property
    def default_fit_prompt_template(self) -> str:
        return self._default_fit_prompt_template

    @property
    def default_interpret_prompt_template(self) -> str:
        return self._default_interpret_prompt_template

    def _run_prompt(self, prompt: str, temperature: float = 0, seed: int | None = None) -> str:
        return run_prompt(
            client=self.client,
            model_name=self.model_name,
            prompt=prompt,
            temperature=temperature,
            seed=seed,
        )

    def fit(
        self,
        X: pd.DataFrame,
        y: np.array,
        temperature: float = 0,
        seed: int | None = None,
        prompt_template: str | None = None,
        prompt_args: dict | None = None,
        try_code: bool = True,
    ) -> None:

        if prompt_template is None:
            prompt_template = self.default_fit_prompt_template

        dataset_str = data_to_text(X, y)

        if prompt_args:
            prompt_args = {"dataset_str": dataset_str, **prompt_args}
        else:
            prompt_args = {"dataset_str": dataset_str, "columns_name": X.columns.tolist() + ["label"]}

        prompt_ = make_prompt(prompt_template=prompt_template, **prompt_args)

        self.code_model = self._run_prompt(prompt=prompt_, seed=seed, temperature=temperature)
        self.code_model = re.sub(r'^```python\n|\n```$', '', self.code_model, flags=re.MULTILINE)
        self.code_model = re.sub(r'^```\n|\n```$', '', self.code_model, flags=re.MULTILINE)

        self.fit_params = dict(temperature=temperature, seed=seed, prompt_template=prompt_template)

        if try_code:
            try:
                self.predict(X.head(1))
                logger.info("Valid code_model successfully created!")
                return self.code_model
            except InvalidCodeModelError:
                raise

    def predict(self, X: pd.DataFrame) -> None:
        if self.code_model is None:
            raise UndefinedCodeModelError("You must load or train the model before predict!")

        _code_space = {}

        try:
            exec(self.code_model, globals(), _code_space)
        except Exception as err:
            logger.exception("Error has occured while prediction")
            raise InvalidCodeModelError("Failed to execute `exec code_model`") from err

        try:
            y = _code_space["predict"](X)
        except Exception as err:
            logger.exception("Error has occured while prediction")
            raise InvalidCodeModelError("Failed to execute `predict` function in code_model") from err

        return y

    def load_code_model(self, file_path: str) -> None:
        with open(file_path) as file:
            self.code_model = file.read()

    def save_code_model(self, file_path: str) -> None:
        if self.code_model is None:
            raise UndefinedCodeModelError("You must train the model before saving!")

        with open(file_path, mode="w") as file:
            file.write(self.code_model)

    def evaluate(self, y_true: np.array, y_pred: np.array) -> dict:
        return evaluate(y_true, y_pred, self.objective)

    def interpret(
        self,
        temperature: float = 0,
        seed: int | None = None,
        prompt_template: str | None = None,
        prompt_args: dict | None = None,
    ) -> None:
        if self.code_model is None:
            raise UndefinedCodeModelError("You must train the model before interpreting!")

        if prompt_template is None:
            prompt_template = self.default_interpret_prompt_template

        if prompt_args is None:
            prompt_args = dict(code_model=self.code_model)

        prompt_ = make_prompt(prompt_template=prompt_template, **prompt_args)

        self.interpret_result = self._run_prompt(prompt=prompt_, temperature=temperature, seed=seed)



from __future__ import annotations

import textwrap

import jinja2
import numpy as np
import pandas as pd


def make_prompt(prompt_template: str, **kwargs) -> str:
    template = jinja2.Template(prompt_template)
    prompt = template.render(**kwargs)
    print(prompt)
    return textwrap.dedent(prompt)


def data_to_text(X: pd.DataFrame, y: np.array):
    df = X.copy()
    df["target"] = y

    data_type = ", ".join(df.dtypes.astype(str))

    dataset = []
    for index, row in df.iterrows():
        row_as_str = [str(item) for item in row.tolist()]
        dataset.append(",".join(row_as_str))
    dataset_str = "\n".join(dataset)
    return dataset_str



import numpy as np
import pandas as pd
from openai import OpenAI
import os
from importlib import resources

from ..utils import preprocessing
from ..utils import openai_model

import warnings
warnings.filterwarnings('ignore')

client = OpenAI()

class IBLModel():
    def __init__(
        self,
        model_name = None,
        prompt_file = None,
        prompt = None,
        seed = None
        ):

        self.model_name = model_name
        self.prompt_file = prompt_file
        self.prompt = prompt
        self.seed = seed
        self.code_model = None


    def _classifier_train(self, X, y):
        dataset_str, data_type = preprocessing._data_to_text(X, y)

        # create prompt
        if self.prompt != None:
            create_prompt = self.prompt.format(
                dataset_str_ = dataset_str,
                data_type_ = data_type
                )

        else:
            if self.prompt_file == 'classification_2.txt':
                with resources.open_text('iblm.iblmodel.prompt', self.prompt_file) as file:
                    self.prompt = file.read()
            elif self.prompt_file == 'classification_3.txt':
                with resources.open_text('iblm.iblmodel.prompt', self.prompt_file) as file:
                    self.prompt = file.read()

            create_prompt = self.prompt.format(
                dataset_str_ = dataset_str,
                data_type_ = data_type
                )

        code_model = openai_model._openai_model(self.model_name, create_prompt, self.seed)

        # prompt modification
        #modification_prompt = preprocessing._prompt_modification(code_model)

        #code_model = openai_model._openai_model(self.model_name, modification_prompt, self.seed)
        self.code_model = code_model

        return self.code_model

    def _predict(self, code, X):
        if self.code_model is None:
            raise Exception("You must train the model before predicting!")

        #code = self.code_model
        exec(code, globals())
        y = predict(X)
        return y



    def _interpret(self):
        if self.code_model is None:
            raise Exception("You must train the model before interpreting!")

        interpret_prompt = preprocessing._interpret_codemodel(self.code_model)
        return openai_model._openai_model(self.model_name, interpret_prompt, self.seed)

    def _generate_python_script(self, file_path, file_name):
        preprocessing._save_codemodel(file_path, file_name, self.code_model)
        return



from sklearn.base import BaseEstimator, ClassifierMixin
from .ibl import IBLModel


class IBLClassifier(BaseEstimator, ClassifierMixin):
    def __init__(
        self,
        model_name = 'gpt-4',
        prompt_file = 'classification_3.txt',
        prompt = None,
        seed = None
        ):
        self.model_name = model_name
        self.prompt_file = prompt_file
        self.prompt = prompt
        self.seed = seed
        self.ibl = IBLModel(self.model_name, self.prompt_file, self.prompt, self.seed)
        self.code_model = None

    def fit(self, X, y):
        self.code_model = self.ibl._classifier_train(X, y)
        return self

    def predict(self, X):
        return self.ibl._predict(self.code_model, X)

    def interpret(self):
        return self.ibl._interpret(self.code_model)

    def generate_python_script(self):
        return self.ibl._generate_python_script(self.code_model)


